package org.dasein.cloud.cloudstack;

import java.io.IOException;
import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.address.AddressServices;
import org.dasein.cloud.services.address.AddressType;
import org.dasein.cloud.services.address.IpForwardingRule;
import org.dasein.cloud.services.firewall.Protocol;
import org.dasein.cloud.services.server.Server;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class IpAddress implements AddressServices {
    static private final String ASSOCIATE_IP_ADDRESS                 = "associateIpAddress";
    static private final String CREATE_PORT_FORWARDING_RULE          = "createPortForwardingRule";
    static private final String DISASSOCIATE_IP_ADDRESS              = "disassociateIpAddress";
    static private final String LIST_PORT_FORWARDING_RULES           = "listPortForwardingRules";
    static private final String LIST_PUBLIC_IP_ADDRESSES             = "listPublicIpAddresses";
    static private final String STOP_FORWARD                         = "deletePortForwardingRule";
    
    private CloudstackProvider provider;
    
    public IpAddress(CloudstackProvider provider) {
        this.provider = provider;
    }
    
    public void assign(String addressId, String toServerId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }
    
    public String forward(String addressId, int publicPort, Protocol protocol, int privatePort, String onServerId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;

        try {
            Server server = provider.getServerServices().getServer(onServerId);
            String privateIpAddress = null;
            
            if( server == null ) {
                throw new CloudException("No such server: " + onServerId);
            }
            for( String addr : server.getPrivateIpAddresses() ) {
                privateIpAddress = addr;
                break;
            }
            if( privateIpAddress == null ) {
                throw new CloudException("Could not determine a private IP address for " + onServerId);
            }
            get.setPath(provider.buildUrl(CREATE_PORT_FORWARDING_RULE, new Param[] { new Param("ipAddress", addressId), new Param("publicPort", String.valueOf(publicPort)), new Param("privatePort", String.valueOf(privatePort)), new Param("protocol", protocol.name()), new Param("virtualMachineId", onServerId)}));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 560 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        } 
        NodeList matches = doc.getElementsByTagName("jobid");
        if( matches.getLength() > 0 ) {
            provider.waitForJob(matches.item(0).getFirstChild().getNodeValue(), "Assigning forwarding rule");
        } 
        for( IpForwardingRule rule : listRules(addressId) ) {
            if( rule.getPublicPort() == publicPort && rule.getPrivatePort() == privatePort && onServerId.equals(rule.getServerId()) && protocol.equals(rule.getProtocol()) && addressId.equals(rule.getAddressId()) ) {
                return rule.getProviderRuleId();
            }
        }
        return null;
    }
    
    public org.dasein.cloud.services.address.IpAddress getIpAddress(String addressId) throws InternalException, CloudException {
        for( org.dasein.cloud.services.address.IpAddress address : listPublicIpPool(false) ) {
            if( address.getProviderIpAddressId().equals(addressId) ) {
                return address;
            }
        }
        return null;
    }
    
    public String getProviderTermForIpAddress(Locale locale) {
        return "IP address";
    }

    public boolean isAssigned(AddressType type) {
        return false;
    }
    
    public boolean isForwarding() {
        return true;
    }

    public boolean isRequestable(AddressType type) {
        return type.equals(AddressType.PUBLIC);
    }
    
    public boolean isSubscribed() {
        return true;
    }
    
    public Iterable<org.dasein.cloud.services.address.IpAddress> listPrivateIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    public Collection<org.dasein.cloud.services.address.IpAddress> listPublicIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            //get.setPath(provider.buildUrl(LIST_PUBLIC_IP_ADDRESSES, new Param[] { new Param("zoneId", provider.getContext().getRegionId()) }));
            get.setPath(provider.buildUrl(LIST_PUBLIC_IP_ADDRESSES, new Param[0]));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 530 || code == 570 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<org.dasein.cloud.services.address.IpAddress> addresses = new ArrayList<org.dasein.cloud.services.address.IpAddress>();
        NodeList matches = doc.getElementsByTagName("publicipaddress");
        for( int i=0; i<matches.getLength(); i++ ) {
            org.dasein.cloud.services.address.IpAddress addr = toAddress(matches.item(i));
            
            if( addr != null && (!unassignedOnly || !addr.isAssigned()) ) {
                addresses.add(addr);
            }
        }
        return addresses;
    }

    public Collection<IpForwardingRule> listRules(String addressId) throws InternalException, CloudException {
        ArrayList<IpForwardingRule> rules = new ArrayList<IpForwardingRule>();
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_PORT_FORWARDING_RULES, new Param[] { new Param("ipAddress", addressId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 560 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        NodeList matches = doc.getElementsByTagName("portforwardingrule");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            if( node != null ) {
                IpForwardingRule rule = new IpForwardingRule();
                NodeList list = node.getChildNodes();

                rule.setAddressId(addressId);
                for( int j=0; j<list.getLength(); j++ ) {
                    Node attr = list.item(j);
                    
                    if( attr.getNodeName().equals("publicport") ) {
                        rule.setPublicPort(Integer.parseInt(attr.getFirstChild().getNodeValue()));                        
                    }
                    else if( attr.getNodeName().equals("privateport") ) {
                        rule.setPrivatePort(Integer.parseInt(attr.getFirstChild().getNodeValue()));                        
                    }
                    else if( attr.getNodeName().equals("protocol") ) {
                        rule.setProtocol(Protocol.valueOf(attr.getFirstChild().getNodeValue().toUpperCase()));
                    } 
                    else if( attr.getNodeName().equals("id") ) {
                        rule.setProviderRuleId(attr.getFirstChild().getNodeValue().toUpperCase());
                    } 
                    else if( attr.getNodeName().equals("virtualmachineid") ) {
                        rule.setServerId(attr.getFirstChild().getNodeValue().toUpperCase());
                    } 
                }
                rules.add(rule);
            }
        }
        return rules;
    }
    
    public void releaseFromPool(String addressId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        int code;
        
        try {
            get.setPath(provider.buildUrl(DISASSOCIATE_IP_ADDRESS, new Param[] { new Param("ipaddress", addressId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 560 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
    }
    
    public void releaseFromServer(String addressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }
    
    public String request(AddressType addressType) throws InternalException, CloudException {
        if( !addressType.equals(AddressType.PUBLIC) ) {
            throw new UnsupportedOperationException("Cannot request private addresses.");
        }
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(ASSOCIATE_IP_ADDRESS, new Param[] { new Param("zoneId", provider.getContext().getRegionId()) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 560 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        NodeList matches = doc.getElementsByTagName("publicipaddress");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            if( node != null ) {
                return toAddress(node).getProviderIpAddressId();
            }
        }
        throw new CloudException("IP address create failed");
    }
    
    public void stopForward(String ruleId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        int code;
        
        try {
            get.setPath(provider.buildUrl(STOP_FORWARD, new Param[] { new Param("id", ruleId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 560 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
    }
    
    private org.dasein.cloud.services.address.IpAddress toAddress(Node node) throws InternalException, CloudException {
        org.dasein.cloud.services.address.IpAddress address = new org.dasein.cloud.services.address.IpAddress();
        NodeList attributes = node.getChildNodes();
        
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node n = attributes.item(i);
            String name = n.getNodeName().toLowerCase();
            String value;
            
            if( n.getChildNodes().getLength() > 0 ) {
                value = n.getFirstChild().getNodeValue();
            }
            else {
                value = null;
            }
            if( name.equals("ipaddress") ) {
                address.setIpAddressId(value);
                address.setAddress(value);
            }
            if( name.equals("zoneid") ) {
                address.setRegionId(value);
            }
        }
        address.setServerId(null);
        if( hasRules(address.getAddress()) ) {
            address.setProviderLoadBalancerId(address.getAddress());
        }
        address.setAddressType(AddressType.PUBLIC);
        return address;
    }
    
    private boolean hasRules(String address) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LoadBalancers.LIST_LOAD_BALANCER_RULES, new Param[] { new Param("publicIp", address) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 431 ) {
                throw new InternalException("Invalid parameters");
            }
            else if( code == 530 || code == 570 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        
        NodeList rules = doc.getElementsByTagName("loadbalancerrule");
        return (rules.getLength() > 0);
    }
}
