package org.dasein.cloud.cloudstack;

import java.io.IOException;
import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;
import java.util.TreeSet;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.image.ImageServices;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Architecture;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.server.Server;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.server.ServerState;
import org.dasein.cloud.services.storage.CloudStoreObject;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class Templates implements ImageServices {
    static private final String CREATE_TEMPLATE             = "createTemplate";
    static private final String DELETE_TEMPLATE             = "deleteTemplate";
    static private final String LIST_OS_TYPES               = "listOsTypes";
    static private final String LIST_TEMPLATES              = "listTemplates";
    static private final String REGISTER_TEMPLATE           = "registerTemplate";
    static private final String UPDATE_TEMPLATE_PERMISSIONS = "updateTemplatePermissions";
    
    private CloudstackProvider provider;
    
    public Templates(CloudstackProvider provider) {
        this.provider = provider;
    }
    
    public String create(String location) throws InternalException, CloudException {
        if( location.startsWith("instance:") ) {
            String name = "image-" + System.currentTimeMillis();
            Server server = provider.getServerServices().getServer(location.substring(9));
            
            if( !server.getCurrentState().equals(ServerState.PAUSED) ) {
                throw new CloudException("The server must be paused in order to create an image.");
            }
            ServerImage image = imageServer(server, name);

            return image.getProviderServerImageId();
        }
        else {
            int idx = location.lastIndexOf("/");
            String name;
            
            if( idx > -1 ) {
                name = location.substring(idx+1);
            }
            else {
                name = location;
            }
            HttpClient client = new HttpClient();
            GetMethod get = new GetMethod();
            Document doc;
            int code;
            
            try {
                Platform platform = Platform.guess(name);
                Architecture architecture = guess(name);
                Param[] params = new Param[8];
                
                params[0] = new Param("name", name);
                params[1] = new Param("displayText", name);
                params[2] = new Param("url", location);
                params[3] = new Param("format", "RAW");
                params[4] = new Param("osTypeId", toOs(platform, architecture));
                params[5] = new Param("zoneId", provider.getContext().getRegionId());
                params[6] = new Param("isPublic", "false");
                params[7] = new Param("isFeatured", "false");
                get.setPath(provider.buildUrl(REGISTER_TEMPLATE, params));
                get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
            }
            catch( SignatureException e ) {
                throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
            }
            try {
                code = client.executeMethod(get);
            }
            catch( HttpException e ) {
                throw new InternalException("HttpException during GET: " + e.getMessage());
            }
            catch( IOException e ) {
                throw new CloudException("IOException during GET: " + e.getMessage());
            }
            if( code != HttpStatus.SC_OK ) {
                if( code == 401 ) {
                    throw new CloudException("Unauthorized user");
                }
                else if( code == 430 ) {
                    throw new InternalException("Malformed parameters");
                }
                else if( code == 547 || code == 530 ) {
                    throw new CloudException("Server error in cloud (" + code + ")");
                }
                else if( code == 531 ) {
                    throw new CloudException("Unable to find account");
                }
                throw new CloudException("Received error code from server: " + code);
            }
            try {
                doc = provider.parseResponse(get.getResponseBodyAsStream());
            }
            catch( IOException e ) {
                throw new CloudException("IOException getting stream: " + e.getMessage());
            }
    
            NodeList matches = doc.getElementsByTagName("templateid");
            String templateId = null;
            
            if( matches.getLength() > 0 ) {
                templateId = matches.item(0).getFirstChild().getNodeValue();
            }
            provider.waitForJob(doc, "Create Template");
            return templateId;
        }
    }

    public void delete(String templateId) throws InternalException, CloudException {
        ServerImage img = getImage(templateId);
        
        if( img == null ) {
            return;
        }
        if( !provider.getContext().getAccountNumber().equals(img.getOwnerAccountId()) ) {
            return;
        }
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(DELETE_TEMPLATE, new Param[] { new Param("id", templateId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        provider.waitForJob(doc, "Delete Template");
    }

    public Collection<ServerImage> find(String executableByAccountNumber, String keyword, Platform platform, String architecture) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            Param[] params;
            
            if( keyword == null ) {
                params = new Param[] { new Param("templateFilter", "executable") };            
            }
            else {
                params = new Param[] { new Param("templateFilter", "executable"), new Param("keyword", keyword) };                            
            }
            String path = provider.buildUrl(LIST_TEMPLATES, params);
            
            get.setPath(path);
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<ServerImage> templates = new ArrayList<ServerImage>();
        NodeList matches = doc.getElementsByTagName("template");
        for( int i=0; i<matches.getLength(); i++ ) {
            ServerImage img = toImage(matches.item(i));
            
            if( img != null ) {
                if( platform != null && !platform.equals(img.getPlatform()) ) {
                    continue;
                }
                if( keyword != null && img.getName().indexOf(keyword) == -1 ) {
                    continue;
                }                
            }
            templates.add(img);
        }
        return templates;
    }

    public ServerImage getImage(String templateId) throws InternalException, CloudException {
        for( ServerImage image : list(provider.getContext().getAccountNumber()) ) {
            if( image.getProviderServerImageId().equals(templateId) ) {
                return image;
            }
        }
        return null;
    }

    public CloudStoreObject getManifest(String templateId) throws CloudException, InternalException {
        return null;
    }

    public String getProviderTermForImage(Locale locale) {
        return "template";
    }

    private String getRootVolume(String serverId) throws InternalException, CloudException {
        return ((Volumes)provider.getVolumeServices()).getRootVolume(serverId);
    }
    
    public Collection<String> getShares(String templateId) throws CloudException, InternalException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_TEMPLATES, new Param[] { new Param("id", templateId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream(), true);
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        TreeSet<String> accounts = new TreeSet<String>();
        NodeList matches = doc.getElementsByTagName("account");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            accounts.add(node.getFirstChild().getNodeValue());
        }
        return accounts;
    }

    private Architecture guess(String desc) {
        Architecture arch = Architecture.I64;
        
        if( desc.contains("x64") ) {
            arch = Architecture.I64;
        }
        else if( desc.contains("x32") ) {
            arch = Architecture.I32;
        }
        else if( desc.contains("64 bit") ) {
            arch = Architecture.I64;
        }
        else if( desc.contains("32 bit") ) {
            arch = Architecture.I32;
        }
        else if( desc.contains("i386") ) {
            arch = Architecture.I32;
        }
        else if( desc.contains("64") ) {
            arch = Architecture.I64;
        }
        else if( desc.contains("32") ) {
            arch = Architecture.I32;
        }
        return arch;
    }
    
    public AsynchronousTask<ServerImage> imageServer(String serverId, String forAccountAccessKey, String directory, String name) throws InternalException, CloudException {
        final Server server = provider.getServerServices().getServer(serverId);
        
        if( !server.getCurrentState().equals(ServerState.PAUSED) ) {
            throw new CloudException("The server must be paused in order to create an image.");
        }
        
        final AsynchronousTask<ServerImage> task = new AsynchronousTask<ServerImage>();
        final String nom = name;
        
        Thread t = new Thread() {
            public void run() {
                try {
                    task.completeWithResult(imageServer(server, nom));
                }
                catch( Throwable t ) {
                    task.complete(t);
                }
            }
        };

        t.setName("Cloud.com Image Building for " + provider.getContext().getAccountNumber());
        t.setDaemon(true);
        t.start();
        return task;
    }
    
    private ServerImage imageServer(Server server, String name) throws CloudException, InternalException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            Param[] params = new Param[8];
            
            params[0] = new Param("name", name);
            params[1] = new Param("displayText", name);
            params[2] = new Param("osTypeId", toOs(server.getPlatform(), server.getArchitecture()));
            params[3] = new Param("zoneId", provider.getContext().getRegionId());
            params[4] = new Param("isPublic", "false");
            params[5] = new Param("isFeatured", "false");
            params[6] = new Param("volumeId", getRootVolume(server.getProviderServerId()));
            params[7] = new Param("passwordEnabled", String.valueOf(isPasswordEnabled(server.getImageId())));
            get.setPath(provider.buildUrl(CREATE_TEMPLATE, params));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream(), true);
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }

        NodeList matches = doc.getElementsByTagName("templateid");
        String templateId = null;
        
        if( matches.getLength() > 0 ) {
            templateId = matches.item(0).getFirstChild().getNodeValue();
        }
        provider.waitForJob(doc, "Create Template");
        if( templateId == null ) {
            throw new CloudException("Failed to provide a template ID.");
        }
        return getImage(templateId);
    }
    
    private boolean isPasswordEnabled(String templateId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            Param[] params = new Param[] { new Param("templateFilter", "executable") };            
            String path = provider.buildUrl(LIST_TEMPLATES, params);
            
            get.setPath(path);
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("template");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            Boolean val = isPasswordEnabled(templateId, node);
            
            if( val != null ) {
                return val.booleanValue();
            }
        }
        return false;
    }
    
    private Boolean isPasswordEnabled(String templateId, Node node) {
        NodeList attributes = node.getChildNodes();
        boolean enabled = false;
        String id = null;
        
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node attribute = attributes.item(i);
            String name = attribute.getNodeName().toLowerCase();
            String value;
            
            if( attribute.getChildNodes().getLength() > 0 ) {
                value = attribute.getFirstChild().getNodeValue();
            }
            else {
                value = null;
            }
            if( name.equalsIgnoreCase("id") ) {
                id = value;
            }
            else if( name.equalsIgnoreCase("passwordenabled") ) {
                enabled = (value != null && value.equalsIgnoreCase("true"));
            }
            if( id != null && enabled ) {
                break;
            }
        } 
        if( id == null || !id.equals(templateId) ) {
            return null;
        }
        return enabled;
    }
    
    public boolean isPublic(String templateId) throws InternalException, CloudException {
        ServerImage image = getImage(templateId);
        
        if( image == null ) {
            return false;
        }
        return image.isPublicImage();
    }

    public boolean isSupportsCustomImages() {
        return true;
    }
    
    public boolean isSupportsSharing() {
        return true;
    }
    
    public Collection<ServerImage> list(String accountNumber) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            Param[] params = new Param[] { new Param("templateFilter", "self") };            
            String path = provider.buildUrl(LIST_TEMPLATES, params);
            
            get.setPath(path);
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<ServerImage> templates = new ArrayList<ServerImage>();
        NodeList matches = doc.getElementsByTagName("template");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            ServerImage image = toImage(node);
            
            if( image != null ) {
                templates.add(image);
            }
        }
        return templates;
    }

    public void share(String templateId, String withAccount, boolean allowed) throws InternalException, CloudException {
        ServerImage img = getImage(templateId);
        
        if( img == null ) {
            return;
        }
        if( !provider.getContext().getAccountNumber().equals(img.getOwnerAccountId()) ) {
            return;
        }
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            if( withAccount == null ) {
                get.setPath(provider.buildUrl(UPDATE_TEMPLATE_PERMISSIONS, new Param[] { new Param("id", templateId), new Param("isPublic", String.valueOf(allowed)) }));
            }
            else {
                String action = (allowed ? "add" : "remove");
                
                get.setPath(provider.buildUrl(UPDATE_TEMPLATE_PERMISSIONS, new Param[] { new Param("id", templateId), new Param("accounts", withAccount), new Param("op", action) }));                
            }
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        provider.waitForJob(doc, "Share Template");
    }
    
    private ServerImage toImage(Node node) throws CloudException, InternalException {
        Architecture bestArchitectureGuess = Architecture.I64;
        NodeList attributes = node.getChildNodes();
        ServerImage image = new ServerImage();
        
        image.setOwnerAccountId(provider.getContext().getAccountNumber());
        image.setType("standard");
        image.setPersistent(true);
        image.setActive(true);
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node attribute = attributes.item(i);
            String name = attribute.getNodeName().toLowerCase();
            String value;
            
            if( attribute.getChildNodes().getLength() > 0 ) {
                value = attribute.getFirstChild().getNodeValue();
            }
            else {
                value = null;
            }
            /*
             * <created></created>
             * <isfeatured>false</isfeatured>
             * <passwordenabled>true</passwordenabled>
             */
            if( name.equals("id") ) {
                image.setProviderServerImageId(value);
            }
            else if( name.equals("zoneid") ) {
                if( !provider.getContext().getRegionId().equals(value) ) {
                    return null;
                }
            }
            else if( name.equalsIgnoreCase("account") ) {
                image.setOwnerAccountId(value);
            }
            else if( name.equals("name") ) {
                image.setName(value);
                if( value != null && value.contains("x64") ) {
                    bestArchitectureGuess = Architecture.I64;
                }
                else if( value != null && value.contains("x32") ) {
                    bestArchitectureGuess = Architecture.I32;
                }
            }
            else if( name.equals("displaytext") ) {
                image.setDescription(value);
                if( value != null && value.contains("x64") ) {
                    bestArchitectureGuess = Architecture.I64;
                }
                else if( value != null && value.contains("x32") ) {
                    bestArchitectureGuess = Architecture.I32;
                }
            }
            else if( name.equals("ispublic") ) {
                image.setPublicImage(value.equalsIgnoreCase("true"));
            }
            else if( name.equals("isfeatured") ) {
                image.setType("featured");
            }
            else if( name.equals("ostypename") ) {
                if( value != null && value.contains("64") ) {
                    bestArchitectureGuess = Architecture.I64;
                }
                else if( value != null && value.contains("32") ) {
                    bestArchitectureGuess = Architecture.I32;
                }
            }
            else if( name.equals("bits") ) {
                if( value.equals("64") ) {
                    image.setArchitecture(Architecture.I64);
                }
                else {
                    image.setArchitecture(Architecture.I32);
                }
            }
            else if( name.equals("created") ) {
                // 2010-06-29T20:49:28+1000
                // TODO: implement when dasein cloud supports template creation timestamps
            }
            else if( name.equals("isready") ) {
                image.setAvailable(value.equalsIgnoreCase("true"));
            }
        }
        Architecture a = image.getArchitecture();
        Collection<ServerSize> sizes = provider.getServerServices().getSupportedSizes(a == null ? Architecture.I64 : a);
        ServerSize[] tmp = new ServerSize[sizes.size()];
        int i = 0;
        for( ServerSize size : sizes ) {
            tmp[i++] = size;
        }
        image.setSize(tmp);
        if( image.getPlatform() == null && image.getName() != null ) {
            image.setPlatform(Platform.guess(image.getName()));
        }
        else {
            image.setPlatform(Platform.UNKNOWN);
        }
        if( image.getArchitecture() == null ) {
            image.setArchitecture(bestArchitectureGuess);
        }
        return image;
    }

    private String toOs(Platform platform, Architecture architecture) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            String path = provider.buildUrl(LIST_OS_TYPES, new Param[0]);
            
            get.setPath(path);
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("ostype");
        for( int i=0; i<matches.getLength(); i++ ) {
            NodeList attrs = matches.item(i).getChildNodes();
            Architecture arch = Architecture.I64;
            Platform pf = null;
            String id = null;
            
            for( int j=0; j<attrs.getLength(); j++ ) {
                Node attr = attrs.item(j);

                if( attr.getNodeName().equals("id") ) {
                    id = attr.getFirstChild().getNodeValue();
                }
                else if( attr.getNodeName().equals("description") ) {
                    String desc = attr.getFirstChild().getNodeValue();
                    
                    pf = Platform.guess(desc);
                    arch = guess(desc);
                }
            }
            if( platform.equals(pf) && architecture.equals(arch) ) {
                return id;
            }
        }
        return null;
    }
}
