package org.dasein.cloud.cloudstack;

import java.io.IOException;
import java.security.SignatureException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.disk.Volume;
import org.dasein.cloud.services.disk.VolumeServices;
import org.dasein.cloud.services.disk.VolumeState;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class Volumes implements VolumeServices {
    static private final String ATTACH_VOLUME = "attachVolume";
    static private final String CREATE_VOLUME = "createVolume";
    static private final String DELETE_VOLUME = "deleteVolume";
    static private final String DETACH_VOLUME = "detachVolume";
    static private final String LIST_DISK_OFFERINGS     = "listDiskOfferings";

    static private final String LIST_VOLUMES  = "listVolumes";

    static public class DiskOffering {
        public String id;
        public long diskSize;
    }
    
    private CloudstackProvider provider;
    
    Volumes(CloudstackProvider provider) {
        this.provider = provider;
    }
    
    @Override
    public void attach(String volumeId, String serverId, String deviceId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Param[] params;
        Document doc;
        int code;
        
        if( deviceId == null ) {
            params = new Param[] { new Param("id", volumeId), new Param("virtualMachineId", serverId) }; 
        }
        else {
            if( deviceId.endsWith("sdh") || deviceId.endsWith("sdh1") ) {
                deviceId = "7";
            }
            else if( deviceId.endsWith("sdi") || deviceId.endsWith("sdh2") || deviceId.endsWith("sdi1") ) {
                deviceId = "8";
            }
            else if( deviceId.endsWith("sdj") || deviceId.endsWith("sdh3") || deviceId.endsWith("sdi2") || deviceId.endsWith("sdj1") ) {
                deviceId = "9";
            }
            else if( deviceId.endsWith("b") || deviceId.endsWith("b1") ) {
                deviceId = "1";
            }
            else if( deviceId.endsWith("c") || deviceId.endsWith("c1") ) {
                deviceId = "2";
            }
            else if( deviceId.endsWith("d") || deviceId.endsWith("d1") ) {
                deviceId = "3";
            }
            else if( deviceId.endsWith("e") || deviceId.endsWith("e1") ) {
                deviceId = "4";
            }
            else if( deviceId.endsWith("f") || deviceId.endsWith("f1") ) {
                deviceId = "5";
            }
            else if( deviceId.endsWith("g") || deviceId.endsWith("g1") ) {
                deviceId = "6";
            }
            else {
                deviceId = "9";
            }
            params = new Param[] { new Param("id", volumeId), new Param("virtualMachineId", serverId), new Param("deviceId", deviceId) };
        }
        try {
            get.setPath(provider.buildUrl(ATTACH_VOLUME, params)); 
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        provider.waitForJob(doc, "Attach Volume");
    }

    @Override
    public String create(String snapshotId, int size, String zoneId) throws InternalException, CloudException {
        String name = "vol_" + System.currentTimeMillis();
        DiskOffering bestOffering = null;
        
        if( snapshotId == null ) {
            for( DiskOffering offering : getDiskOfferings() ) {
                if( offering.diskSize == size ) {
                    bestOffering = offering;
                    break;
                }
                else if( bestOffering == null ) {
                    bestOffering = offering;
                }
                else if( bestOffering.diskSize < size ) {
                    if( offering.diskSize > size || offering.diskSize > bestOffering.diskSize ) {
                        bestOffering = offering;
                    }
                }
                else if( offering.diskSize > size && offering.diskSize < bestOffering.diskSize ) {
                    bestOffering = offering;
                }
            }
            if( bestOffering == null ) {
                throw new CloudException("No matching volume size offering for " + size);
            }
        }
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            Param[] params = new Param[3];

            params[0] = new Param("name", name);
            params[1] = new Param("zoneId", provider.getContext().getRegionId());
            if( snapshotId != null ) {
                params[2] = new Param("snapshotId", snapshotId);
            }
            else {
                params[2] = new Param("diskOfferingId", bestOffering.id);
            }
            get.setPath(provider.buildUrl(CREATE_VOLUME, params));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }

        NodeList matches = doc.getElementsByTagName("volumeid");
        String volumeId = null;
        
        if( matches.getLength() > 0 ) {
            volumeId = matches.item(0).getFirstChild().getNodeValue();
        }
        provider.waitForJob(doc, "Create Volume");
        return volumeId;
    }

    @Override
    public void delete(String volumeId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(DELETE_VOLUME, new Param[] { new Param("id", volumeId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        provider.waitForJob(doc, "Delete Volume");
    }

    @Override
    public void detach(String volumeId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(DETACH_VOLUME, new Param[] { new Param("id", volumeId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        provider.waitForJob(doc, "Detach Volume");
    }

    Collection<DiskOffering> getDiskOfferings() throws InternalException, CloudException {
        ArrayList<DiskOffering> offerings = new ArrayList<DiskOffering>();
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_DISK_OFFERINGS, new Param[0]));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("diskoffering");
        for( int i=0; i<matches.getLength(); i++ ) {
            DiskOffering offering = new DiskOffering();
            Node node = matches.item(i);
            NodeList attributes;
            
            attributes = node.getChildNodes();
            for( int j=0; j<attributes.getLength(); j++ ) {
                Node n = attributes.item(j);
                String value;

                if( n.getChildNodes().getLength() > 0 ) {
                    value = n.getFirstChild().getNodeValue();
                }
                else {
                    value = null;
                }
                if( n.getNodeName().equals("id") ) {
                    offering.id = value;
                }
                else if( n.getNodeName().equals("disksize") ) {
                    offering.diskSize = Long.parseLong(value);
                }
                if( offering.id != null && offering.diskSize > 0 ) {
                    break;
                }
            }
            if( offering.id != null ) {
                offerings.add(offering);
            }
        }
        return offerings;
    }
    
    @Override
    public String getProviderTermForVolume(Locale locale) {
        return "volume";
    }

    String getRootVolume(String serverId) throws InternalException, CloudException {
        for( Volume volume : list(true) ) {
            if( volume.getServerId() != null && volume.getServerId().equals(serverId) ) {
                return volume.getProviderVolumeId();
            }
        }
        return null;
    }
    
    @Override
    public org.dasein.cloud.services.disk.Volume getVolume(String volumeId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            try {
                Long.parseLong(volumeId);
            }
            catch( NumberFormatException e ) {
                return null;
            }
            get.setPath(provider.buildUrl(LIST_VOLUMES, new Param[] { new Param("id", volumeId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("volume");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node v = matches.item(i);

            if( v != null ) {
                org.dasein.cloud.services.disk.Volume volume = toVolume(v, false);
                
                if( volume != null ) {
                    return volume;
                }
            }
        }
        return null;
    }

    @Override
    public Collection<org.dasein.cloud.services.disk.Volume> list() throws InternalException, CloudException {
        return list(false);
    }
     
    private Collection<Volume> list(boolean rootOnly) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_VOLUMES, new Param[0]));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 547 || code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            else if( code == 531 ) {
                throw new CloudException("Unable to find account");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<org.dasein.cloud.services.disk.Volume> volumes = new ArrayList<org.dasein.cloud.services.disk.Volume>();
        NodeList matches = doc.getElementsByTagName("volume");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node v = matches.item(i);

            if( v != null ) {
                org.dasein.cloud.services.disk.Volume volume = toVolume(v, rootOnly);
                
                if( volume != null ) {
                    volumes.add(volume);
                }
            }
        }
        return volumes;
    }

    private org.dasein.cloud.services.disk.Volume toVolume(Node node, boolean rootOnly) {
        if( node == null ) {
            return null;
        }
        
        org.dasein.cloud.services.disk.Volume volume = new org.dasein.cloud.services.disk.Volume();
        NodeList attributes = node.getChildNodes();
        boolean root = false;
        
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node attribute = attributes.item(i);
            
            if( attribute != null ) {
                String name = attribute.getNodeName();
                
                if( name.equals("id") ) {
                    volume.setProviderVolumeId(attribute.getFirstChild().getNodeValue());
                }
                if( name.equals("zoneid") ) {
                    if( !provider.getContext().getRegionId().equals(attribute.getFirstChild().getNodeValue()) ) {
                        return null;
                    }
                }
                else if( name.equals("type") && attribute.hasChildNodes() ) {
                    if( attribute.getFirstChild().getNodeValue().equalsIgnoreCase("root") ) {
                        root = true;
                    }
                }
                else if( name.equals("name") && attribute.hasChildNodes() ) {
                    if( volume.getName() == null ) {
                        volume.setName(attribute.getFirstChild().getNodeValue());
                    }
                }
                else if( name.equalsIgnoreCase("virtualmachineid") && attribute.hasChildNodes() ) {
                    volume.setServerId(attribute.getFirstChild().getNodeValue());
                }
                else if( name.equals("displayname") && attribute.hasChildNodes() ) {
                    volume.setName(attribute.getFirstChild().getNodeValue());
                }
                else if( name.equals("size") && attribute.hasChildNodes() ) {
                    volume.setSizeInGigabytes((int)(Long.parseLong(attribute.getFirstChild().getNodeValue())/1024000000L));
                }
                else if( name.equals("state") && attribute.hasChildNodes() ) {
                    String state = attribute.getFirstChild().getNodeValue();
                    
                    if( state == null ) {
                        volume.setCurrentState(VolumeState.PENDING);
                    }
                    else if( state.equalsIgnoreCase("created") ) {
                        volume.setCurrentState(VolumeState.AVAILABLE);
                    }
                    else {
                        volume.setCurrentState(VolumeState.PENDING);
                    }
                }
                else if( name.equals("created") && attribute.hasChildNodes() ) {
                    String date = attribute.getFirstChild().getNodeValue();
                    DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ"); //2009-02-03T05:26:32.612278
                    
                    try {
                        volume.setCreationTimestamp(df.parse(date).getTime());
                    }
                    catch( ParseException e ) {
                        volume.setCreationTimestamp(0L);
                    }
                }
            }
        }
        if( !root && rootOnly ) {
            return null;
        }
        String name = volume.getName();
        
        if( name == null ) {
            name = volume.getProviderVolumeId();
            volume.setName(name);
        }
        volume.setRegionId(provider.getContext().getRegionId());
        volume.setDataCenterId(Zones.getDataCenterForRegion(provider.getContext().getRegionId()));
        if( root ) {
            volume.setDeviceId("/dev/sda2");
        }
        else {
            volume.setDeviceId("/dev/xvdh");
        }
        return volume;
    }
}
