package org.dasein.cloud.cloudstack;

import java.io.IOException;
import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Locale;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.dc.DataCenter;
import org.dasein.cloud.services.dc.DataCenterServices;
import org.dasein.cloud.services.dc.Region;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class Zones implements DataCenterServices {
    static private final String LIST_ZONES = "listZones";
    
    static public final String getDataCenterForRegion(String regionId) {
        return (regionId + "-Main");
    }
    
    private CloudstackProvider provider;
    
    public Zones(CloudstackProvider provider) {
        this.provider = provider;
    }
    
    public DataCenter getDataCenter(String zoneId) throws InternalException, CloudException {
        for( Region region : listRegions() ) {
            for( DataCenter dc : listDataCenters(region.getProviderRegionId()) ) {
                if( dc.getProviderDataCenterId().equals(zoneId) ) {
                    return dc;
                }
            }
        }
        return null;
    }

    public String getProviderTermForDataCenter(Locale locale) {
        return "zone";
    }

    public String getProviderTermForRegion(Locale locale) {
        return "region";
    }

    public Region getRegion(String regionId) throws InternalException, CloudException {
        for( Region region : listRegions() ) {
            if( region.getProviderRegionId().equals(regionId) ) {
                return region;
            }
        }
        return null;
    }

    private transient HashMap<String,Collection<DataCenter>> dataCenters = new HashMap<String,Collection<DataCenter>>();
    
    public Collection<DataCenter> listDataCenters(String regionId) throws InternalException, CloudException {
        Collection<DataCenter> results = dataCenters.get(regionId);
        
        if( results == null ) {
            DataCenter zone = new DataCenter();
            
            zone.setActive(true);
            zone.setAvailable(true);
            zone.setName(getDataCenterForRegion(regionId));
            zone.setProviderDataCenterId(getDataCenterForRegion(regionId));
            zone.setRegionId(regionId);
            results = new ArrayList<DataCenter>();
            results.add(zone);
            dataCenters.put(regionId, results);
        }
        return results;
    }

    public Collection<Region> listRegions() throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_ZONES, new Param[] { new Param("available", "true") }));
            get.addRequestHeader("Content-Type", "application/x-www-form-urlencoded; charset=utf-8");
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        if( code != HttpStatus.SC_OK ) {
            try {
                provider.parseResponse(get.getResponseBodyAsStream(), true);
            }
            catch( Throwable ignore ) {
            }
            if( code == 401 ) {
                throw new CloudException("Unauthorized user");
            }
            else if( code == 430 ) {
                throw new InternalException("Malformed parameters");
            }
            else if( code == 530 ) {
                throw new CloudException("Server error in cloud (" + code + ")");
            }
            throw new CloudException("Received error code from server: " + code);
        }
        try {
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<Region> regions = new ArrayList<Region>();
        NodeList matches = doc.getElementsByTagName("zone");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            regions.add(toRegion(node));
        }
        return regions;
    }

    public String matchDataCenter(CloudProvider provider, String zoneId) throws InternalException, CloudException {
        return zoneId;
    }
    
    private Region toRegion(Node node) {
        NodeList attributes = node.getChildNodes();
        Region region = new Region();
        
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node n = attributes.item(i);
            String name = n.getNodeName().toLowerCase();
            String value;
            
            if( n.getChildNodes().getLength() > 0 ) {
                value = n.getFirstChild().getNodeValue();
            }
            else {
                value = null;
            }
            if( name.equals("id") ) {
                region.setProviderRegionId(value);
            }
            else if( name.equals("name") ) {
                region.setName(value);
            }
        }
        region.setActive(true);
        region.setAvailable(true);
        return region;
    }
}
